/* $Id: crires_combine.c,v 1.36 2012-09-05 14:14:17 yjung Exp $
 *
 * This file is part of the CRIRES Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: yjung $
 * $Date: 2012-09-05 14:14:17 $
 * $Revision: 1.36 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
   								Includes
 -----------------------------------------------------------------------------*/

#include <math.h>
#include <cpl.h>

#include "irplib_spectrum.h"

#include "crires_combine.h"
#include "crires_pfits.h"
#include "crires_load.h"
#include "crires_utils.h"

/*-----------------------------------------------------------------------------
   								Functions prototypes
 -----------------------------------------------------------------------------*/

static cpl_image * crires_combine_collapse_sigclip(const cpl_imagelist *, 
        const cpl_vector *, double, int) ;
static cpl_image ** crires_combine_chip(const cpl_frameset *, 
        const cpl_frameset *, crires_illum_period, int, const char *, 
        const char *, const char *, const char *,  int, int, int, int, int, 
        double) ;
static cpl_image ** crires_combine_chip_nodded(const cpl_imagelist *,
        const cpl_vector *, int, int, int) ;
static cpl_image ** crires_combine_chip_staring(const cpl_imagelist *,
        cpl_vector *, int) ;
static cpl_vector * crires_combine_get_offsets(const cpl_frameset *) ;
static cpl_vector * crires_combine_get_offsets_blind(cpl_imagelist *) ;
static int * crires_combine_classif(const cpl_vector *, int *) ;
static int crires_identify_ab(const cpl_vector *, int) ;
static cpl_imagelist * crires_util_saa_groups(const cpl_imagelist *,
        const cpl_vector *, const int *, int, cpl_vector **) ;
static cpl_imagelist * crires_util_nodded(cpl_imagelist *, cpl_vector *,
        cpl_vector **) ;
static double crires_util_refine_offset(const cpl_image *, const cpl_image *) ;
static int off_comp(double, double, double) ;
static cpl_image ** crires_combine_chip_win(const cpl_frameset *, int, 
        const char *, const char *, const char *, const char *, int, int, 
        int, double) ;

/*----------------------------------------------------------------------------*/
/**
 * @defgroup crires_combine     Images combination routines
 *
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    The image combination routine --> imagelist
  @param    rawframes   the spectrum frames
  @param    skyframes   the sky frames
  @param    period
  @param    flat        the flat frame or NULL
  @param    dark        the dark frame or NULL
  @param    bpm         the bpm frame or NULL
  @param    detlin      the detlin frame or NULL
  @param    nodding     the flag for nodding
  @param    blind       the flag to search blindly the offsets
  @param    refine      the flag to refine the offsets
  @param    onlyA       Also combine only NOD A images
  @param    onlyB       Also combine only NOD B images
  @return   NULL in error case. Otherwise, 6 image lists:
            - the combined images of the 4 detectors in ADU/sec
            - the contribution maps of the 4 detectors
            - the combined images in ADU/sec of Nodded A of the 4 detectors 
            or NULL if not requested
            - the contribution maps of Nodded A of the 4 detectors or NULL 
            if not requested
            - the combined images in ADU/sec of Nodded B of the 4 detectors 
            or NULL if not requested
            - the contribution maps of Nodded B of the 4 detectors or NULL 
            if not requested
 */
/*----------------------------------------------------------------------------*/
cpl_imagelist ** crires_combine_imagelist(
        const cpl_frameset  *   rawframes,
        const cpl_frameset  *   skyframes,
        crires_illum_period     period,
        const char          *   flat,
        const char          *   dark,
        const char          *   bpm,
        const char          *   detlin,
        int                     nodding,
        int                     blind,
        int                     refine,
        int                     onlyA,
        int                     onlyB)
{
    const cpl_frame     *   ref_frame ;
    cpl_propertylist    *   plist ;
    double                  dit ;
    cpl_imagelist       **  out ;
    cpl_image           **  imas ;
    int                     i ;

    /* Test entries */
    if (rawframes== NULL) return NULL ;
    if (period == CRIRES_ILLUM_UNKNOWN) return NULL ;

    /* Get the DIT value from the header */
    ref_frame = cpl_frameset_get_position_const(rawframes, 0) ;
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                    0)) == NULL) {
        cpl_msg_error(__func__, "getting header from frame");
        return NULL ;
    }
    dit = crires_pfits_get_dit(plist) ;
    cpl_propertylist_delete(plist) ;
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "Cannot get the DIT from the header") ;
        return NULL ;
    }
   
    /* Create output imagelist */
    out = cpl_malloc(6 * sizeof(cpl_imagelist*)) ;
    out[0] = cpl_imagelist_new() ;
    out[1] = cpl_imagelist_new() ;
    if (onlyA) {
        /* Combined Nodded A images of the 4 detectors */
        out[2] = cpl_imagelist_new() ;
        /* Contribution maps Nodded A of the 4 detectors */
        out[3] = cpl_imagelist_new() ;
    } else {
        out[2] = NULL ;
        out[3] = NULL ;
    }
     if (onlyB) {
        /* Combined Nodded B images of the 4 detectors */
        out[4] = cpl_imagelist_new() ;
        /* Contribution maps Nodded B of the 4 detectors */
        out[5] = cpl_imagelist_new() ;
    } else {
        out[4] = NULL ;
        out[5] = NULL ;
    }

    /* Loop on the chips */
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        /* Combine for current chip */
        if ((imas = crires_combine_chip(rawframes, skyframes, period, i+1, 
                        flat, dark, bpm, detlin, nodding,
                        refine, blind, onlyA, onlyB, dit)) == NULL) {
            cpl_msg_error(__func__, "Cannot combine chip %d", i+1) ;
            cpl_imagelist_delete(out[0]) ;
            cpl_imagelist_delete(out[1]) ;
            cpl_free(out) ;
            return NULL ;
        }
        cpl_imagelist_set(out[0], imas[0], i) ;
        cpl_imagelist_set(out[1], imas[1], i) ;
        if (onlyA) {
            cpl_imagelist_set(out[2], imas[2], i) ;
            cpl_imagelist_set(out[3], imas[3], i) ;
        }
        if (onlyB) {
            cpl_imagelist_set(out[4], imas[4], i) ;
            cpl_imagelist_set(out[5], imas[5], i) ;
        }
        cpl_free(imas) ;
    }
    return out ;
}
/*----------------------------------------------------------------------------*/
/**
  @brief    The image combination routine for win mode --> imagelist
  @param    rawframes   the spectrum frames
  @param    flat        the flat frame or NULL
  @param    dark        the dark frame or NULL
  @param    bpm         the bpm frame or NULL
  @param    detlin      the detlin frame or NULL
  @param    refine      the flag to refine the offsets
  @param    onlyA       Also combine only NOD A images
  @param    onlyB       Also combine only NOD B images
  @return   NULL in error case. Otherwise, 6 image lists:
            - the combined images of the 4 detectors in ADU/sec
            - the contribution maps of the 4 detectors
            - the combined images in ADU/sec of Nodded A of the 4 detectors 
            or NULL if not requested
            - the contribution maps of Nodded A of the 4 detectors or NULL 
            if not requested
            - the combined images in ADU/sec of Nodded B of the 4 detectors 
            or NULL if not requested
            - the contribution maps of Nodded B of the 4 detectors or NULL 
            if not requested
 */
/*----------------------------------------------------------------------------*/
cpl_imagelist ** crires_combine_imagelist_win(
        const cpl_frameset  *   rawframes,
        const char          *   flat,
        const char          *   dark,
        const char          *   bpm,
        const char          *   detlin,
        int                     refine,
        int                     onlyA,
        int                     onlyB)
{
    const cpl_frame     *   ref_frame ;
    cpl_propertylist    *   plist ;
    double                  dit ;
    cpl_imagelist       **  out ;
    cpl_image           **  images_chip2 ;
    cpl_image           **  images_chip3 ;
    cpl_image           *   empty_image ;

    /* Test entries */
    if (rawframes== NULL) return NULL ;

    /* Get the DIT value from the header */
    ref_frame = cpl_frameset_get_position_const(rawframes, 0) ;
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                    0)) == NULL) {
        cpl_msg_error(__func__, "getting header from frame");
        return NULL ;
    }
    dit = crires_pfits_get_dit(plist) ;
    cpl_propertylist_delete(plist) ;
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "Cannot get the DIT from the header") ;
        return NULL ;
    }

    /* Combine Detector 2 */
    if ((images_chip2 = crires_combine_chip_win(rawframes,
                    2, flat, dark, bpm, detlin, refine, onlyA, onlyB, 
                    dit)) == NULL) {
        cpl_msg_error(__func__, "Cannot combine chip 2") ;
        return NULL ;
    }

    /* Combine Detector 3 */
    if ((images_chip3 = crires_combine_chip_win(rawframes,
                    3, flat, dark, bpm, detlin, refine, onlyA, onlyB, 
                    dit)) == NULL) {
        cpl_msg_error(__func__, "Cannot combine chip 3") ;
        cpl_image_delete(images_chip2[0]) ;
        cpl_image_delete(images_chip2[1]) ;
        if (onlyA) cpl_image_delete(images_chip2[2]) ;
        if (onlyA) cpl_image_delete(images_chip2[3]) ;
        if (onlyB) cpl_image_delete(images_chip2[4]) ;
        if (onlyB) cpl_image_delete(images_chip2[5]) ;
        cpl_free(images_chip2) ;
        return NULL ;
    }

    /* Create output imagelist */
    out = cpl_malloc(6 * sizeof(cpl_imagelist*)) ;
    /* Combined images of the 4 detectors */
    out[0] = cpl_imagelist_new() ;
    /* Contribution maps of the 4 detectors */
    out[1] = cpl_imagelist_new() ;
    if (onlyA) {
        /* Combined Nodded A images of the 4 detectors */
        out[2] = cpl_imagelist_new() ;
        /* Contribution maps Nodded A of the 4 detectors */
        out[3] = cpl_imagelist_new() ;
    } else {
        out[2] = NULL ;
        out[3] = NULL ;
    }
     if (onlyB) {
        /* Combined Nodded B images of the 4 detectors */
        out[4] = cpl_imagelist_new() ;
        /* Contribution maps Nodded B of the 4 detectors */
        out[5] = cpl_imagelist_new() ;
    } else {
        out[4] = NULL ;
        out[5] = NULL ;
    }

    /* Create the combined images list */
    empty_image = cpl_image_duplicate(images_chip2[0]) ;
    cpl_image_multiply_scalar(empty_image, 0.0) ;
    cpl_imagelist_set(out[0], cpl_image_duplicate(empty_image), 0);
    cpl_imagelist_set(out[0], images_chip2[0], 1) ;
    cpl_imagelist_set(out[0], images_chip3[0], 2) ;
    cpl_imagelist_set(out[0], cpl_image_duplicate(empty_image), 3);
    cpl_image_delete(empty_image) ;

    /* Create the contribution maps list */
    empty_image = cpl_image_duplicate(images_chip2[1]) ;
    cpl_image_multiply_scalar(empty_image, 0.0) ;
    cpl_imagelist_set(out[1], cpl_image_duplicate(empty_image), 0);
    cpl_imagelist_set(out[1], images_chip2[1], 1) ;
    cpl_imagelist_set(out[1], images_chip3[1], 2) ;
    cpl_imagelist_set(out[1], cpl_image_duplicate(empty_image), 3);
    cpl_image_delete(empty_image) ;

    if (onlyA) {
        /* Create the Nodded A combined images list */
        empty_image = cpl_image_duplicate(images_chip2[2]) ;
        cpl_image_multiply_scalar(empty_image, 0.0) ;
        cpl_imagelist_set(out[2], cpl_image_duplicate(empty_image), 0);
        cpl_imagelist_set(out[2], images_chip2[2], 1) ;
        cpl_imagelist_set(out[2], images_chip3[2], 2) ;
        cpl_imagelist_set(out[2], cpl_image_duplicate(empty_image), 3);
        cpl_image_delete(empty_image) ;

        /* Create the Nodded A contribution maps list */
        empty_image = cpl_image_duplicate(images_chip2[3]) ;
        cpl_image_multiply_scalar(empty_image, 0.0) ;
        cpl_imagelist_set(out[3], cpl_image_duplicate(empty_image), 0);
        cpl_imagelist_set(out[3], images_chip2[3], 1) ;
        cpl_imagelist_set(out[3], images_chip3[3], 2) ;
        cpl_imagelist_set(out[3], cpl_image_duplicate(empty_image), 3);
        cpl_image_delete(empty_image) ;
    }

    if (onlyB) {
        /* Create the Nodded B combined images list */
        empty_image = cpl_image_duplicate(images_chip2[4]) ;
        cpl_image_multiply_scalar(empty_image, 0.0) ;
        cpl_imagelist_set(out[4], cpl_image_duplicate(empty_image), 0);
        cpl_imagelist_set(out[4], images_chip2[4], 1) ;
        cpl_imagelist_set(out[4], images_chip3[4], 2) ;
        cpl_imagelist_set(out[4], cpl_image_duplicate(empty_image), 3);
        cpl_image_delete(empty_image) ;

        /* Create the Nodded B contribution maps list */
        empty_image = cpl_image_duplicate(images_chip2[5]) ;
        cpl_image_multiply_scalar(empty_image, 0.0) ;
        cpl_imagelist_set(out[5], cpl_image_duplicate(empty_image), 0);
        cpl_imagelist_set(out[5], images_chip2[5], 1) ;
        cpl_imagelist_set(out[5], images_chip3[5], 2) ;
        cpl_imagelist_set(out[5], cpl_image_duplicate(empty_image), 3);
        cpl_image_delete(empty_image) ;
    }
    cpl_free(images_chip2) ;
    cpl_free(images_chip3) ;

    return out ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The images calibration routine for a given chip
  @param    ilist       the list of frames to calibrate 
  @param    period
  @param    chip        the chip to calibrate (1 to CRIRES_NB_DETECTORS)
  @param    flat        the flat frame or NULL
  @param    dark        the dark frame or NULL
  @param    bpm         the bpm frame or NULL
  @param    detlin      the detlin  frame or NULL
  @param    dit         the DIT for the dark correction
  @param    stripe_center_y     the y center position of the lower stripe
  @param    stripe_offset_y     the y offset between the 2 stripes centers
  @param    stripe_size_y       the y size of a stripe
  The flat, dark and bpm must have the same size as the input ilist. If WIN 
  mode, the data must match.
  In the case of detlin, data are only taken in normal mode. Ifthe data to 
  calibrate are WIN data, the stripe_* parameters tell where the detlin areas
  must be extracted from the passed full 1024x512 detlin data. 
  If stripe_* parameters are negative (normal mode), the full detlin data are 
  taken.

  @return   0 if ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
int crires_calib_chip_list(
        cpl_imagelist       *   ilist,
        crires_illum_period     period,
        int                     chip,
        const char          *   flat,
        const char          *   dark,
        const char          *   bpm,
        const char          *   detlin,
        double                  dit,
        int                     stripe_center_y,
        int                     stripe_offset_y,
        int                     stripe_size_y)
{
    cpl_imagelist       *   callist ;
    int                     i ;
   
    /* Test entries */
    if (ilist == NULL) return -1 ;
    if (period == CRIRES_ILLUM_UNKNOWN) return -1 ;
    if (chip < 1 || chip > CRIRES_NB_DETECTORS) return -1 ;
 
    /* Clean the bad pixels */
    if (bpm != NULL) {
        cpl_msg_info(__func__, "Correct the bad pixels") ;
        for (i=0 ; i<cpl_imagelist_get_size(ilist) ; i++) {
            if (crires_bpm_correct_image(cpl_imagelist_get(ilist, i),
                        period, bpm, chip) != 0) {
                cpl_msg_error(__func__, 
                        "Cannot clean the bad pixels in obj %d", i+1);
                return -1 ;
            }
        }
    }

    /* Apply the dark */
    if (dark != NULL) {
        cpl_msg_info(__func__, "Correct for the dark") ; 
        if ((callist = crires_load_file(dark, period, 
                        CPL_TYPE_FLOAT)) == NULL) {
            cpl_msg_error(__func__, "Cannot load the dark") ;
            return -1 ;
        }
        /* Multiply the dark by dit */
        cpl_imagelist_multiply_scalar(callist, dit) ;
        /* Subtract the dark */
        if (cpl_imagelist_subtract_image(ilist, 
                    cpl_imagelist_get(callist, chip-1))!=CPL_ERROR_NONE) {
            cpl_msg_error(__func__, "Cannot apply the dark") ;
            cpl_imagelist_delete(callist) ;
            return -1 ;
        }
        cpl_imagelist_delete(callist) ;
    }
  
    /* Apply the non linearity correction */
    if (detlin != NULL) {
        cpl_msg_info(__func__, "Correct for the Non-Linearity") ; 
        if (crires_detlin_correct(ilist, period, detlin, chip,
                    stripe_center_y, stripe_offset_y, stripe_size_y)) {
            cpl_msg_error(__func__, "Cannot correct for the Non-Linearity") ;
            return -1 ;
        }
    }
    
    /* Apply the flatfield */
    if (flat != NULL) {
        cpl_msg_info(__func__, "Correct for the flat field") ; 
        /* Load the flat */
        if ((callist = crires_load_file(flat, period, CPL_TYPE_FLOAT))==NULL) {
            cpl_msg_error(__func__, "Cannot load the flat field") ;
            return -1 ;
        }
        /* Divide */
        if (cpl_imagelist_divide_image(ilist, 
                    cpl_imagelist_get(callist, chip-1))!=CPL_ERROR_NONE) {
            cpl_msg_error(__func__, "Cannot apply the flat field") ;
            cpl_imagelist_delete(callist) ;
            return -1 ;
        }
        cpl_imagelist_delete(callist) ;
    }
    return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Collapse an image list (several methods)
  @param    ilist       the input image list
  @param    norm_fact   the normalisation factors
  @param    kappa       the kappa value for the sigma clipping
  @param    chip        the chip to recombine (1 to CRIRES_NB_DETECTORS)
  @param    meth        the requested method
  @return   The collapsed image or NULL in error case
  The supported methods are CRIRES_COLLAPSE_AVER, CRIRES_COLLAPSE_MEDIAN,
  CRIRES_COLLAPSE_SIG
 */
/*----------------------------------------------------------------------------*/
cpl_image * crires_combine_collapse_imagelist(
        const cpl_imagelist     *   ilist,
        const cpl_vector        *   norm_fact,
        double                      kappa,
        int                         chip,
        crires_collapse_method      meth)
{
    cpl_image               *   ima ;
    crires_collapse_method      meth_loc ;

    /* Test entries */
    if (ilist == NULL) return NULL ;
    if (chip < 1 || chip > CRIRES_NB_DETECTORS) return NULL ;
    if (norm_fact == NULL) return NULL ;
    
    /* Test number of frames */
    if (cpl_imagelist_get_size(ilist) < 3) {
        cpl_msg_warning(__func__, "Not enough frames - apply average") ;
        meth_loc = CRIRES_COLLAPSE_AVG ;
    } else {
        meth_loc = meth ;
    }

    /* Switch on the method */
    switch (meth_loc) {
        case CRIRES_COLLAPSE_AVG:
            if ((ima = cpl_imagelist_collapse_create(ilist)) == NULL) {
                cpl_msg_error(__func__, "Cannot average") ;
                return NULL ;
            }
            break ;

        case CRIRES_COLLAPSE_MED:
            if ((ima = cpl_imagelist_collapse_median_create(ilist)) == NULL) {
                cpl_msg_error(__func__, "Cannot median") ;
                return NULL ;
            }
            break ;

        case CRIRES_COLLAPSE_SIG:
            if ((ima = crires_combine_collapse_sigclip(ilist, norm_fact,
                            kappa, chip)) == NULL) {
                cpl_msg_error(__func__, 
                        "Cannot apply the sigma-clipping method") ;
                return NULL ;
            }
            break ;
        default:
            cpl_msg_error(__func__, "Unsupported mode") ;
            return NULL ;
    }
    return ima ;
}

/**@}*/

/*----------------------------------------------------------------------------*/
/**
  @brief    The Sigma clipping combination method
  @param    ilist       the input image list
  @param    norm_fact   the normalisation factors
  @param    kappa       the kappa value for the sigma clipping
  @param    chip        the chip to recombine (1 to CRIRES_NB_DETECTORS)
  @return   the collapsed image or NULL in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_image * crires_combine_collapse_sigclip(
        const cpl_imagelist     *   ilist,
        const cpl_vector        *   norm_fact,
        double                      kappa,
        int                         chip)
{
    double          gain, ron_e ;
    double          min_norm, sum ;
    cpl_image   *   medians ;
    cpl_image   *   exp_stddev ;
    cpl_image   *   threshold ;
    cpl_image   *   out ;
    float       *   pout ;
    const float *   pilist ;
    float       *   pthreshold ;
    int             nvals, nx, ny, nima ;
    int             i, j, k ;

    /* Test entries */
    if (ilist == NULL) return NULL ;
    if (norm_fact == NULL) return NULL ;
    if (chip < 1 || chip > CRIRES_NB_DETECTORS) return NULL ;

    /* Initialise */
    nima = cpl_imagelist_get_size(ilist) ;
    nx = cpl_image_get_size_x(cpl_imagelist_get_const(ilist, 0)) ;
    ny = cpl_image_get_size_y(cpl_imagelist_get_const(ilist, 0)) ;

    /* Get gain and ron_e */
    if (chip == 1) { 
        gain = CRIRES_GAIN_CHIP1 ;
        ron_e = CRIRES_RONE_CHIP1 ;
    }
    if (chip == 2) { 
        gain = CRIRES_GAIN_CHIP2 ;
        ron_e = CRIRES_RONE_CHIP2 ;
    }
    if (chip == 3) { 
        gain = CRIRES_GAIN_CHIP3 ;
        ron_e = CRIRES_RONE_CHIP3 ;
    }
    if (chip == 4) { 
        gain = CRIRES_GAIN_CHIP4 ;
        ron_e = CRIRES_RONE_CHIP4 ;
    }

    /* Compute minimum of the norm_fact */
    min_norm = cpl_vector_get_min(norm_fact) ;

    /* Compute medians */
    if ((medians = cpl_imagelist_collapse_median_create(ilist)) == NULL) {
        cpl_msg_error(__func__, "Cannot median") ;
        return NULL ;
    }

    /* Compute exp_stddev */
    exp_stddev = cpl_image_multiply_scalar_create(medians, gain) ;
    cpl_image_add_scalar(exp_stddev, ron_e*ron_e) ;
    /* To fix some rare negative numbers due to some side effects */
    cpl_image_threshold(exp_stddev, 0.0, DBL_MAX, 0.0, 0.0) ;
    cpl_image_power(exp_stddev, 0.5) ;
    if (fabs(gain*min_norm)) 
        cpl_image_divide_scalar(exp_stddev, gain*min_norm) ;

    /* Create threshold image */
    threshold = cpl_image_multiply_scalar_create(exp_stddev, kappa) ;
    cpl_image_add(threshold, medians) ;
    cpl_image_delete(exp_stddev) ;
    cpl_image_delete(medians) ;

    /* Create output image */
    out = cpl_image_duplicate(threshold) ;
    cpl_image_multiply_scalar(out, 0.0) ;

    /* Loop on the pixels to create the output image */
    pout = cpl_image_get_data_float(out) ;
    pthreshold = cpl_image_get_data_float(threshold) ;
    for (j=0 ; j<ny ; j++) {
        for (i=0 ; i< nx ; i++) {
            nvals = 0 ;
            sum = 0.0 ;
            for (k=0 ; k< nima ; k++) {
                pilist=cpl_image_get_data_float_const(
                        cpl_imagelist_get_const(ilist, k)) ;
                if (pilist[i+j*nx] < pthreshold[i+j*nx]) {
                    nvals ++ ;
                    sum += pilist[i+j*nx] ;
                }
            }
            if (nvals > 0)  pout[i+j*nx] = sum / nvals ;
            else            pout[i+j*nx] = 0.0 ;
        }
    }
    cpl_image_delete(threshold) ;

    return out ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The image combination routine for a given chip in win mode
  @param    rawframes   the spectrum frames
  @param    chip        the chip to recombine (1 to CRIRES_NB_DETECTORS)
  @param    flat        the flat frame or NULL
  @param    dark        the dark frame or NULL
  @param    bpm         the bpm frame or NULL
  @param    detlin      the detlin frame or NULL
  @param    refine      the flag to refine the offsets
  @param    onlyA       Also combine only NOD A images
  @param    onlyB       Also combine only NOD B images
  @param    dit         the DIT in sec.
  @return   NULL in error case. Otherwise, 6 images:
            - the combined image in ADU/sec
            - the contribution map 
            - the combined image in ADU/sec of Nodded A or NULL if not requested
            - the contribution map of Nodded A or NULL if not requested
            - the combined image in ADU/sec of Nodded B or NULL if not requested
            - the contribution map of Nodded B or NULL if not requested
 */
/*----------------------------------------------------------------------------*/
static cpl_image ** crires_combine_chip_win(
        const cpl_frameset  *   rawframes,
        int                     chip,
        const char          *   flat,
        const char          *   dark,
        const char          *   bpm,
        const char          *   detlin,
        int                     refine,
        int                     onlyA,
        int                     onlyB,
        double                  dit)
{
    cpl_imagelist       *   ilist ;
    cpl_imagelist       *   ilist_extended ;
    cpl_vector          *   offsets_vec ;
    cpl_image           **  out ;
    cpl_image           **  out_final ;
    const char          *   fname ;
    cpl_propertylist    *   plist ;
    int                     offset_y, stripe_y, center_y ;

    /* Test entries */
    if (rawframes == NULL) return NULL ;
    if (chip < 2 || chip > CRIRES_NB_DETECTORS-1) return NULL ;
    if (dit <= 0) return NULL ;

    /* Load the images */
    cpl_msg_info(__func__, "Load the chip number %d", chip) ; 
    cpl_msg_indent_more() ;
    if ((ilist = crires_load_frameset(rawframes, CRIRES_ILLUM_FULL_DETECTOR, 
                    chip, CPL_TYPE_FLOAT))==NULL) {
        cpl_msg_error(__func__, "Cannot load the data") ;
        cpl_msg_indent_less() ;
        return NULL ;
    }
    cpl_msg_indent_less() ;

    /* Get the stripes related informations */
    fname = cpl_frame_get_filename(
            cpl_frameset_get_position_const(rawframes,0)) ;
    if ((plist=cpl_propertylist_load(fname, 0)) == NULL) return NULL ;
    offset_y = crires_pfits_get_stripe_offsety(plist) ;
    stripe_y = (int)(cpl_image_get_size_y(cpl_imagelist_get(ilist, 0))/2.0) ;
    center_y = crires_pfits_get_stripe_centery(plist) ;
    cpl_propertylist_delete(plist) ;

    /* Apply the calibrations */
    if (crires_calib_chip_list(ilist, CRIRES_ILLUM_FULL_DETECTOR, chip, flat, 
                dark, bpm, detlin, dit, center_y, offset_y, stripe_y)) {
        cpl_msg_error(__func__, "Cannot apply the calibrations") ;
        cpl_imagelist_delete(ilist) ;
        return NULL ;
    }

    /* Get the offsets from the header */
    cpl_msg_info(__func__, "Get the offsets from the header") ; 
    if ((offsets_vec = crires_combine_get_offsets(rawframes)) == NULL) {
        cpl_msg_error(__func__, "Cannot get the offsets from the header");
        cpl_imagelist_delete(ilist) ;
        return NULL ;
    }

    /* Reconstruct the image list with the gap between the stripes */
    if ((ilist_extended = crires_util_ilist_extend(ilist, offset_y)) == NULL) {
        cpl_msg_error(__func__, "Cannot extend the image list") ;
        cpl_imagelist_delete(ilist) ;
        cpl_vector_delete(offsets_vec) ;
        return NULL ;
    }
    cpl_imagelist_delete(ilist) ;

    /* Combination of the frames */
    cpl_msg_info(__func__, "Recombine the frames") ; 
    cpl_msg_indent_more() ;
    out = crires_combine_chip_nodded(ilist_extended, offsets_vec,
            refine, onlyA, onlyB) ;
    cpl_vector_delete(offsets_vec) ;
    cpl_imagelist_delete(ilist_extended) ;
    cpl_msg_indent_less() ;
    if (out == NULL) {
        cpl_msg_error(__func__, "Cannot combine the images") ;
        return NULL ;
    }
    
    /* Convert from ADU to ADU/sec */
    cpl_image_divide_scalar(out[0], dit) ;
    if (onlyA) cpl_image_divide_scalar(out[2], dit) ;
    if (onlyB) cpl_image_divide_scalar(out[4], dit) ;

    /* Remove the introduced gap from out */
    out_final = cpl_malloc(6*sizeof(cpl_image*)) ;
    out_final[0] = crires_util_remove_gap(out[0], stripe_y) ;
    cpl_image_delete(out[0]) ;
    out_final[1] = crires_util_remove_gap(out[1], stripe_y) ;
    cpl_image_delete(out[1]) ;

    if (onlyA) {
        out_final[2] = crires_util_remove_gap(out[2], stripe_y) ;
        cpl_image_delete(out[2]) ;
        out_final[3] = crires_util_remove_gap(out[3], stripe_y) ;
        cpl_image_delete(out[3]) ;
    } else {
        out_final[2] = NULL ;
        out_final[3] = NULL ;
    }
    if (onlyB) {
        out_final[4] = crires_util_remove_gap(out[4], stripe_y) ;
        cpl_image_delete(out[4]) ;
        out_final[5] = crires_util_remove_gap(out[5], stripe_y) ;
        cpl_image_delete(out[5]) ;
    } else {
        out_final[4] = NULL ;
        out_final[5] = NULL ;
    }
    cpl_free(out) ;

    return out_final ;
}


/*----------------------------------------------------------------------------*/
/**
  @brief    The image combination routine for a given chip
  @param    rawframes   the spectrum frames
  @param    skyframes   the sky frames
  @param    period
  @param    chip        the chip to recombine (1 to CRIRES_NB_DETECTORS)
  @param    flat        the flat frame or NULL
  @param    dark        the dark frame or NULL
  @param    bpm         the bpm frame or NULL
  @param    detlin      the detlin  frame or NULL
  @param    nodding     the flag for nodding
  @param    refine      the flag to refine the offsets
  @param    blind       the flag to search blindly the offsets
  @param    onlyA       Also combine only NOD A images
  @param    onlyB       Also combine only NOD B images
  @param    dit         the DIT in sec.
  @return   NULL in error case. Otherwise, 6 images:
            - the combined image in ADU/sec
            - the contribution map 
            - the combined image in ADU/sec of Nodded A or NULL if not requested
            - the contribution map of Nodded A or NULL if not requested
            - the combined image in ADU/sec of Nodded B or NULL if not requested
            - the contribution map of Nodded B or NULL if not requested
 */
/*----------------------------------------------------------------------------*/
static cpl_image ** crires_combine_chip(
        const cpl_frameset  *   rawframes,
        const cpl_frameset  *   skyframes,
        crires_illum_period     period,
        int                     chip,
        const char          *   flat,
        const char          *   dark,
        const char          *   bpm,
        const char          *   detlin,
        int                     nodding,
        int                     refine,
        int                     blind,
        int                     onlyA,
        int                     onlyB,
        double                  dit)
{
    cpl_imagelist       *   ilist ;
    cpl_imagelist       *   skylist ;
    cpl_image           *   sky_ima ;
    cpl_vector          *   offsets_vec ;
    cpl_image           **  out ;
   
    /* Test entries */
    if (rawframes == NULL) return NULL ;
    if (period == CRIRES_ILLUM_UNKNOWN) return NULL ;
    if (chip < 1 || chip > CRIRES_NB_DETECTORS) return NULL ;
    if (dit <= 0) return NULL ;

    /* Load the images */
    cpl_msg_info(__func__, "Load the chip number %d", chip) ; 
    cpl_msg_indent_more() ;
    if ((ilist = crires_load_frameset(rawframes, period, chip, 
                    CPL_TYPE_FLOAT))==NULL) {
        cpl_msg_error(__func__, "Cannot load the data") ;
        cpl_msg_indent_less() ;
        return NULL ;
    }
    cpl_msg_indent_less() ;

    /* Subtract the sky if present */
    if (skyframes != NULL) {
        /* Load the sky */
        if ((skylist = crires_load_frameset(skyframes, period, chip, 
                        CPL_TYPE_FLOAT))==NULL) {
            cpl_msg_warning(__func__, "Cannot load the sky frames") ;
        } else {
            if (cpl_frameset_get_size(skyframes) ==
                    cpl_frameset_get_size(rawframes)){
                /* Subtract the sky frames */
                cpl_msg_info(__func__, "Subtract the sky frames") ; 
                if (cpl_imagelist_subtract(ilist, skylist) != CPL_ERROR_NONE){
                    cpl_msg_warning(__func__, "Cannot subtract the sky") ;
                }
            } else {
                /* Note: This is useless in nodding  */
                cpl_msg_warning(__func__, 
                        "Number of sky frames does not match the objects") ;
                cpl_msg_warning(__func__,
                        "Collapse the sky frames together for correction") ;
                sky_ima = cpl_imagelist_collapse_create(skylist) ;
                /* Subtract the sky frame */
                if (cpl_imagelist_subtract_image(ilist, 
                            sky_ima) != CPL_ERROR_NONE){
                    cpl_msg_warning(__func__, "Cannot subtract the sky") ;
                }
                cpl_image_delete(sky_ima) ;
            }
            cpl_imagelist_delete(skylist) ;
        }
    }

    /* Apply the calibrations */
    if (crires_calib_chip_list(ilist, period, chip, flat, dark, bpm, detlin, 
                dit, -1, -1, -1)) {
        cpl_msg_error(__func__, "Cannot apply the calibrations") ;
        cpl_imagelist_delete(ilist) ;
        return NULL ;
    }

    /* Get the offsets */
    if (blind) {
        /* Get the offsets blindly */
        cpl_msg_info(__func__, "Get the offsets blindly") ; 
        if ((offsets_vec=crires_combine_get_offsets_blind(ilist)) == NULL) {
            cpl_msg_error(__func__, "Cannot get the offsets blindly") ;
            cpl_imagelist_delete(ilist) ;
            return NULL ;
        }
    } else {
        /* Get the offsets from the header */
        cpl_msg_info(__func__, "Get the offsets from the header") ; 
        if ((offsets_vec = crires_combine_get_offsets(rawframes)) == NULL) {
            cpl_msg_error(__func__, "Cannot get the offsets from the header");
            cpl_imagelist_delete(ilist) ;
            return NULL ;
        }
    }

    /* Combination of the frames */
    cpl_msg_info(__func__, "Recombine the frames") ; 
    cpl_msg_indent_more() ;
    if (nodding) {
        out = crires_combine_chip_nodded(ilist, offsets_vec, refine,
                onlyA, onlyB) ;
    } else {
        out = crires_combine_chip_staring(ilist, offsets_vec, refine) ;
    }
    cpl_vector_delete(offsets_vec) ;
    cpl_imagelist_delete(ilist) ;
    cpl_msg_indent_less() ;
    if (out == NULL) {
        cpl_msg_error(__func__, "Cannot combine the images") ;
        return NULL ;
    }

    /* Convert from ADU to ADU/sec */
    cpl_image_divide_scalar(out[0], dit) ;
    if (onlyA) cpl_image_divide_scalar(out[2], dit) ;
    if (onlyB) cpl_image_divide_scalar(out[4], dit) ;
    return out ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The image combination in staring
  @param    ilist       the images to combine
  @param    offsets_vec the associated offsets
  @param    refine      the flag to refine the offsets
  @return   the combined image and the contribution map or NULL
 */
/*----------------------------------------------------------------------------*/
static cpl_image ** crires_combine_chip_staring(
        const cpl_imagelist *   ilist,
        cpl_vector          *   offsets_vec,
        int                     refine)
{
    cpl_vector          *   off_y ;
    double              *   poff_y ;
    cpl_vector          *   off_x ;
    cpl_bivector        *   offsets ;
    cpl_image           **  combined ;
    double                  new_offset ;
    int                     i ;
    
    /* Test entries */
    if (ilist == NULL) return NULL ;
    if (offsets_vec == NULL) return NULL ;
    
    /* Create local offsets */
    off_y = cpl_vector_duplicate(offsets_vec) ;

    /* Subtract the first offset to the others */
    cpl_vector_subtract_scalar(off_y, cpl_vector_get(off_y,0)) ;
    
    /* Refine the offsets if requested */
    if (refine) {
        cpl_msg_info(__func__, "Refine the offsets") ;
        poff_y = cpl_vector_get_data(offsets_vec) ;
        for (i=0 ; i<cpl_imagelist_get_size(ilist) ; i++) {
            new_offset = crires_util_refine_offset(
                    cpl_imagelist_get_const(ilist, 0),
                    cpl_imagelist_get_const(ilist, i)) ;
            if (new_offset < 5000) {
                if (fabs(new_offset-poff_y[i]) < 10) {
                    cpl_msg_info(__func__, "Offset %d: %g --> %g", 
                            i+1, poff_y[i], new_offset);
                    poff_y[i] = new_offset ;
                } else {
                    cpl_msg_info(__func__, "Offset %d: %g --> Not refined",
                            i+1, poff_y[i]) ; 
                }
            } else {
                cpl_msg_info(__func__, "Offset %d: %g --> Not refined", 
                        i+1, poff_y[i]) ; 
            }
        }
    }

    /* Images combination */
    /* Get the offsets in a bivector */
    off_x = cpl_vector_duplicate(off_y) ;
    cpl_vector_fill(off_x, 0.0) ;
    offsets = cpl_bivector_wrap_vectors(off_x, off_y) ;
    /* Shift and add */
    cpl_msg_info(__func__, "Shift and Add the nodded frames") ;
    if ((combined = cpl_geom_img_offset_saa(ilist, offsets,
                    CPL_KERNEL_DEFAULT, 0, 0, CPL_GEOM_FIRST,
                    NULL, NULL)) == NULL) {
        cpl_msg_error(__func__, "Cannot shift and add group") ;
        cpl_bivector_unwrap_vectors(offsets) ;
        cpl_vector_delete(off_x) ;
        cpl_vector_delete(off_y) ;
        return NULL ;
    }
    cpl_bivector_unwrap_vectors(offsets) ;
    cpl_vector_delete(off_x) ;
    cpl_vector_delete(off_y) ;

    return combined ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The image combination in nodding
  @param    ilist       the images to combine
  @param    offsets_vec the associated offsets
  @param    refine      the flag to refine the offsets
  @param    onlyA       Also combine only NOD A images
  @param    onlyB       Also combine only NOD B images
  @return   NULL in error case. Otherwise, 6 images:
            - the combined image 
            - the contribution map 
            - the combined image of Nodded A or NULL if not requested
            - the contribution map of Nodded A or NULL if not requested
            - the combined image of Nodded B or NULL if not requested
            - the contribution map of Nodded B or NULL if not requested
 */
/*----------------------------------------------------------------------------*/
static cpl_image ** crires_combine_chip_nodded(
        const cpl_imagelist *   ilist,
        const cpl_vector    *   offsets_vec,
        int                     refine,
        int                     onlyA,
        int                     onlyB)
{
    int                 *   groups ;
    int                     ngroups ;
    int                     nima ;
    cpl_imagelist       *   abba ;
    cpl_vector          *   abba_off ;
    cpl_imagelist       *   nodded ;
    cpl_imagelist       *   nodded_onlyA ;
    cpl_imagelist       *   nodded_onlyB ;
    cpl_vector          *   nodded_off_y ;
    cpl_vector          *   nodded_off_y_onlyA ;
    cpl_vector          *   nodded_off_y_onlyB ;
    double              *   pnodded_off_y ;
    cpl_vector          *   nodded_off_x ;
    cpl_bivector        *   nodded_offsets ;
    cpl_image           **  out ;
    cpl_image           **  combined ;
    cpl_image           **  combined_onlyA ;
    cpl_image           **  combined_onlyB ;
    double                  new_offset ;
    int                     i, j ;
    
    /* Test entries */
    if (ilist == NULL) return NULL ;
    if (offsets_vec == NULL) return NULL ;
    
    /* Classify in groups the a and b images sequence */
    cpl_msg_info(__func__, "Classify in groups") ; 
    cpl_msg_indent_more() ;
    if ((groups = crires_combine_classif(offsets_vec, &ngroups)) == NULL) {
        cpl_msg_error(__func__, "Cannot classify the data") ;
        cpl_msg_indent_less() ;
        return NULL ;
    }
    cpl_msg_indent_less() ;
  
    /* Shift and add each group to one image */
    cpl_msg_info(__func__, "Shift and add each group to one image") ; 
    cpl_msg_indent_more() ;
    if ((abba = crires_util_saa_groups(ilist, offsets_vec, groups, 
                    ngroups, &abba_off)) == NULL) {
        cpl_msg_error(__func__, "cannot shift and add groups") ;
        cpl_free(groups) ;
        cpl_msg_indent_less() ;
        return NULL ;
    }
    cpl_free(groups) ;
    cpl_msg_indent_less() ;
 
   /* Create the nodded images */
    cpl_msg_info(__func__, "Create the nodded images") ; 
    cpl_msg_indent_more() ;
    if ((nodded = crires_util_nodded(abba, abba_off, 
                    &nodded_off_y))==NULL) {
        cpl_msg_error(__func__, "cannot create the nodded images") ;
        cpl_imagelist_delete(abba) ;
        cpl_vector_delete(abba_off) ;
        cpl_msg_indent_less() ;
        return NULL ;
    }
    cpl_imagelist_delete(abba) ;
    cpl_vector_delete(abba_off) ;
    cpl_msg_indent_less() ;
   
    /* Refine the offsets if requested */
    if (refine) {
        cpl_msg_info(__func__, "Refine the offsets") ;
        pnodded_off_y = cpl_vector_get_data(nodded_off_y) ;
        for (i=0 ; i<cpl_imagelist_get_size(nodded) ; i++) {
            new_offset = crires_util_refine_offset(
                    cpl_imagelist_get(nodded, 0),
                    cpl_imagelist_get(nodded, i)) ;
             if (new_offset < 5000) {
                if (fabs(new_offset-pnodded_off_y[i]) < 10) {
                    cpl_msg_info(__func__, "Offset %d: %g --> %g", 
                            i+1, pnodded_off_y[i], new_offset);
                    pnodded_off_y[i] = new_offset ;
                } else {
                    cpl_msg_info(__func__, "Offset %d: %g --> Not refined",
                            i+1, pnodded_off_y[i]) ; 
                }
            } else {
                cpl_msg_info(__func__, "Offset %d: %g --> Not refined", 
                        i+1, pnodded_off_y[i]) ; 
            }
        }
    }

    /* Combine only Nodded A images together */
    if (onlyA) {
        /* Count how many Nodded A images */
        nima = 0 ;
        for (i=0 ; i<cpl_imagelist_get_size(nodded) ; i++) {
            if (crires_identify_ab(nodded_off_y, i) == 1) nima ++ ;
        }

        /* Extract the nodded A offsets and images */
        nodded_off_y_onlyA = cpl_vector_new(nima) ;
        nodded_onlyA = cpl_imagelist_new() ;
        j = 0 ; 
        for (i=0 ; i<cpl_imagelist_get_size(nodded) ; i++) {
            /* Is the current offset a Nodded A one ? */
            if (crires_identify_ab(nodded_off_y, i) == 1) {
                /* nodded_off_y -> nodded_off_y_onlyA */
                cpl_vector_set(nodded_off_y_onlyA, j, 
                        cpl_vector_get(nodded_off_y, i)) ;
                /* nodded -> nodded_onlyA */
                cpl_imagelist_set(nodded_onlyA, 
                        cpl_image_duplicate(cpl_imagelist_get(nodded, i)), j) ;
                j++ ;
            }
        }

        nodded_off_x = cpl_vector_duplicate(nodded_off_y_onlyA) ;
        cpl_vector_fill(nodded_off_x, 0.0) ;
        nodded_offsets = cpl_bivector_wrap_vectors(nodded_off_x, 
                nodded_off_y_onlyA) ;
        /* Shift and add */
        cpl_msg_info(__func__, "Shift and Add the A nodded frames") ;
        if ((combined_onlyA = cpl_geom_img_offset_saa(nodded_onlyA, 
                        nodded_offsets,
                        CPL_KERNEL_DEFAULT, 0, 0, CPL_GEOM_FIRST, NULL, 
                        NULL)) == NULL) {
            cpl_msg_error(__func__, "Cannot shift and add nodded A") ;
            cpl_imagelist_delete(nodded_onlyA) ;
            cpl_bivector_unwrap_vectors(nodded_offsets) ;
            cpl_vector_delete(nodded_off_x) ;
            cpl_vector_delete(nodded_off_y_onlyA) ;
            cpl_vector_delete(nodded_off_y) ;
            cpl_imagelist_delete(nodded) ;
            return NULL ;
        }
        cpl_imagelist_delete(nodded_onlyA) ;
        cpl_bivector_unwrap_vectors(nodded_offsets) ;
        cpl_vector_delete(nodded_off_x) ;
        cpl_vector_delete(nodded_off_y_onlyA) ;
    } else {
        combined_onlyA = NULL ;
    } 

    /* Combine only Nodded B images together */
    if (onlyB) {
        /* Count how many Nodded B images */
        nima = 0 ;
        for (i=0 ; i<cpl_imagelist_get_size(nodded) ; i++) {
            if (crires_identify_ab(nodded_off_y, i) == 0) nima ++ ;
        }

        /* Extract the nodded B offsets and images */
        nodded_off_y_onlyB = cpl_vector_new(nima) ;
        nodded_onlyB = cpl_imagelist_new() ;
        j = 0 ; 
        for (i=0 ; i<cpl_imagelist_get_size(nodded) ; i++) {
            /* Is the current offset a Nodded B one ? */
            if (crires_identify_ab(nodded_off_y, i) == 0) {
                /* nodded_off_y -> nodded_off_y_onlyB */
                cpl_vector_set(nodded_off_y_onlyB, j, 
                        cpl_vector_get(nodded_off_y, i)) ;
                /* nodded -> nodded_onlyB */
                cpl_imagelist_set(nodded_onlyB, 
                        cpl_image_duplicate(cpl_imagelist_get(nodded, i)), j) ;
                j++ ;
            }
        }

        /* Subtract the first offset to the others */
        cpl_vector_subtract_scalar(nodded_off_y_onlyB, 
                cpl_vector_get(nodded_off_y_onlyB,0)) ;

        nodded_off_x = cpl_vector_duplicate(nodded_off_y_onlyB) ;
        cpl_vector_fill(nodded_off_x, 0.0) ;
        nodded_offsets = cpl_bivector_wrap_vectors(nodded_off_x, 
                nodded_off_y_onlyB) ;
        /* Shift and add */
        cpl_msg_info(__func__, "Shift and Add the B nodded frames") ;
        if ((combined_onlyB = cpl_geom_img_offset_saa(nodded_onlyB, 
                        nodded_offsets,
                        CPL_KERNEL_DEFAULT, 0, 0, CPL_GEOM_FIRST, NULL, 
                        NULL)) == NULL) {
            cpl_msg_error(__func__, "Cannot shift and add nodded B") ;
            cpl_imagelist_delete(nodded_onlyB) ;
            cpl_bivector_unwrap_vectors(nodded_offsets) ;
            cpl_vector_delete(nodded_off_x) ;
            cpl_vector_delete(nodded_off_y_onlyB) ;
            cpl_vector_delete(nodded_off_y) ;
            cpl_imagelist_delete(nodded) ;
            return NULL ;
        }
        cpl_imagelist_delete(nodded_onlyB) ;
        cpl_bivector_unwrap_vectors(nodded_offsets) ;
        cpl_vector_delete(nodded_off_x) ;
        cpl_vector_delete(nodded_off_y_onlyB) ;
    } else {
        combined_onlyB = NULL ;
    } 

    /* Images combination */
    /* Get the offsets in a bivector */
    nodded_off_x = cpl_vector_duplicate(nodded_off_y) ;
    cpl_vector_fill(nodded_off_x, 0.0) ;
    nodded_offsets = cpl_bivector_wrap_vectors(nodded_off_x, nodded_off_y) ;
    /* Shift and add */
    cpl_msg_info(__func__, "Shift and Add the nodded frames") ;
    if ((combined = cpl_geom_img_offset_saa(nodded, nodded_offsets,
                    CPL_KERNEL_DEFAULT, 0, 0, CPL_GEOM_FIRST, NULL, 
                    NULL)) == NULL) {
        cpl_msg_error(__func__, "Cannot shift and add nodded") ;
        cpl_imagelist_delete(nodded) ;
        cpl_bivector_unwrap_vectors(nodded_offsets) ;
        cpl_vector_delete(nodded_off_x) ;
        cpl_vector_delete(nodded_off_y) ;
        return NULL ;
    }
    cpl_imagelist_delete(nodded) ;
    cpl_bivector_unwrap_vectors(nodded_offsets) ;
    cpl_vector_delete(nodded_off_x) ;
    cpl_vector_delete(nodded_off_y) ;

    /* Return out */
    out = cpl_malloc(6*sizeof(cpl_image*)) ;
    out[0] = combined[0] ;
    out[1] = combined[1] ;
    cpl_free(combined) ;
    if (onlyA) {
        out[2] = combined_onlyA[0] ;
        out[3] = combined_onlyA[1] ;
        cpl_free(combined_onlyA) ;
    } else {
        out[2] = NULL ;
        out[3] = NULL ;
    }
    if (onlyB) {
        out[4] = combined_onlyB[0] ;
        out[5] = combined_onlyB[1] ;
        cpl_free(combined_onlyB) ;
    } else {
        out[4] = NULL ;
        out[5] = NULL ;
    }
    return out ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the offsets
  @param    rawframes   the input rawframes
  @return   the offsets in a vector
 */
/*----------------------------------------------------------------------------*/
static cpl_vector * crires_combine_get_offsets(const cpl_frameset * rawframes)
{
    cpl_vector          *   offsets ;
    double              *   pvect ;
    int                     nraw ;
    const cpl_frame     *   cur_frame ;
    cpl_propertylist    *   plist ;
    int                     i ;

    /* Test entries */
    if (rawframes == NULL) return NULL ;
    
    /* Initialise */
    nraw = cpl_frameset_get_size(rawframes) ;

    /* Get the rawframes Y offsets */
    offsets = cpl_vector_new(nraw) ;
    pvect = cpl_vector_get_data(offsets) ;
    for (i=0 ; i<nraw ; i++) {
        cur_frame = cpl_frameset_get_position_const(rawframes, i) ;
        if ((plist=cpl_propertylist_load(cpl_frame_get_filename(cur_frame),
                        0)) == NULL) {
            cpl_msg_error(__func__, "cannot get property list") ;
            cpl_vector_delete(offsets) ;
            return NULL ;
        }
        pvect[i] = crires_pfits_get_cumoffsety(plist) ;
        if (cpl_error_get_code()) {
            cpl_msg_error(__func__, "Cannot get the cumulative Y offset") ;
            cpl_vector_delete(offsets) ;
            cpl_propertylist_delete(plist) ;
            return NULL ;
        }
        cpl_propertylist_delete(plist) ;
    }
    return offsets ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the offsets from a rough spectrum detection
  @param    ilist       the input image list
  @return   the offsets in a vector
 */
/*----------------------------------------------------------------------------*/
static cpl_vector * crires_combine_get_offsets_blind(cpl_imagelist * ilist)
{
    cpl_vector          *   offsets ;
    double              *   pvect ;
    int                     nraw ;
    int                     i ;

    /* Test entries */
    if (ilist == NULL) return NULL ;
   
    /* Initialise */
    nraw = cpl_imagelist_get_size(ilist) ;

    /* Create the offset vector */
    offsets = cpl_vector_new(nraw) ;
    pvect = cpl_vector_get_data(offsets) ;
    for (i=0 ; i<nraw ; i++) {
        pvect[i] = crires_util_refine_offset(
                cpl_imagelist_get(ilist, 0),
                cpl_imagelist_get(ilist, i)) ;
    }
    return offsets ;
}
   
 
/*----------------------------------------------------------------------------*/
/**
  @brief    Identify if an offset is in the A or B 'group'.
  @param    offsets     the offsets
  @param    index       the index to identify (0->n-1)
  @return   0 if the required offset is in the A group, 1 in the B, -1
            in error case
  A group is all offsets that are close to the first offset, B group are
  the others.

    Example:
        Offset 0 = -252     -> A group by definition    (return 0)
        Offset 1 = +10      -> B group                  (return 1)
        Offset 2 = +12      -> B group                  (return 1)
        Offset 3 = -220     -> A group                  (return 0)
        Offset 4 = -260     -> A group                  (return 0)
        Offset 5 = +30      -> B group                  (return 1)
 */
/*----------------------------------------------------------------------------*/
static int crires_identify_ab(
        const cpl_vector    *   offsets,
        int                     index)
{
    int                     noffs ;
    cpl_vector          *   tmp_vec ;
    const double        *   pvect ;
    double                  offset_thresh ;

    /* Test entries */
    if (offsets == NULL) return -1 ;

    /* Initialise */
    noffs = cpl_vector_get_size(offsets) ;

    /* Test entries */
    if (index >= noffs) return -1 ;

    /* Separate the offsets in 2 categories */
    tmp_vec = cpl_vector_duplicate(offsets) ;
    cpl_vector_sort(tmp_vec, 1) ;
    pvect = cpl_vector_get_data_const(tmp_vec) ;
    if (pvect[0] == pvect[noffs-1]) {
        cpl_msg_error(__func__, "Only one offset in the list - abort") ;
        cpl_vector_delete(tmp_vec) ;
        return -1 ;
    }
    offset_thresh = (pvect[0] + pvect[noffs-1]) / 2.0 ;
    cpl_vector_delete(tmp_vec) ;

    /* Identify the different A and B groups */
    pvect = cpl_vector_get_data_const(offsets) ;
    
    /* Test on which side is the required offset */
    if ((pvect[index] - offset_thresh) * (pvect[0] - offset_thresh) > 0) {
        return 1 ;
    } else { 
        return 0 ;
    }
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The abba classification
  @param    offsets     the images offsets
  @param    ngroups     the computed number of groups (must be even)
  @return   the groups array
 
  The classification is done according the offsets read in header of the 
  rawframes.
  
  The frames are first classified in two different categories A or B depending 
  in which zone their offset is.

  Example
  frame1 - offset: -250.4    ---> cat. A
  frame2 - offset: -120.6    ---> cat. B
  frame3 - offset: -150.6    ---> cat. B
  frame4 - offset: -220.4    ---> cat. A

  These frames are then grouped in cubes according the category order.
  Each group is then averaged in a single image.

  Example 
  frame1 -  cat A    -> group 1  
  frame2 -  cat A    -> group 1
  frame3 -  cat B    -> group 2 
  frame4 -  cat B    -> group 2
  frame5 -  cat B    -> group 3
  frame6 -  cat B    -> group 3
  frame7 -  cat A    -> group 4
  frame8 -  cat A    -> group 4
  frame9 -  cat A    -> group 5
  frame10 - cat A    -> group 5
  frame11 - cat B    -> group 6
  frame12 - cat B    -> group 6
  frame13 - cat B    -> group 6
  frame14 - cat B    -> group 6
 */
/*----------------------------------------------------------------------------*/
static int * crires_combine_classif(
        const cpl_vector    *   offsets,
        int                 *   ngroups)
{
    const double        *   pvect ;
    int                     nraw ;
    double                  offset_thresh ;
    cpl_vector          *   tmp_vec ;
    int                 *   groups ;
    int                     last_group ;
    int                     i, j, k, l ;

    /* Test entries */
    if (offsets == NULL) return NULL ;

    /* Initialise */
    nraw = cpl_vector_get_size(offsets) ;

    /* Separate the offsets in 2 categories */
    tmp_vec = cpl_vector_duplicate(offsets) ;
    cpl_vector_sort(tmp_vec, 1) ;
    pvect = cpl_vector_get_data_const(tmp_vec) ;
    if (pvect[0] == pvect[nraw-1]) {
        cpl_msg_error(__func__, "Only one offset in the list - abort") ;
        cpl_vector_delete(tmp_vec) ;
        return NULL ;
    }
    offset_thresh = (pvect[0] + pvect[nraw-1]) / 2.0 ;
    cpl_vector_delete(tmp_vec) ;

    /* Identify the different A and B groups */
    pvect = cpl_vector_get_data_const(offsets) ;
    *ngroups = 0 ;
    groups = cpl_calloc(nraw, sizeof(int)) ;

    /* Check if some offsets are ambiguous */
    for (i=0 ; i<nraw ; i++) {
        if (offset_thresh == pvect[i])
            cpl_msg_warning(__func__, "Offset nb %d is amiguous", i+1) ;
    }

    /* Create a look up table to associate the ith obj with the jth frame */
    i = 0 ;
    while (i < nraw) {
        j = 0 ;
        /* Count the number of successive '+' or '-' (j) */
        while ((i+j<nraw) &&
                (!off_comp(pvect[i], pvect[i+j], offset_thresh))) j++ ;

        if (i+j >= nraw) i = nraw ;
        else {
            k = 0 ;
            /* Check if there are j '-' or '+' (k) */
            while ((i+j+k < nraw)
                    && (!off_comp(pvect[i+j], pvect[i+j+k], offset_thresh))
                    && (k<j)) k++ ;
            last_group = 1 ;
            if (i+j+k < nraw) {
                for (l=i+j+k ; l<nraw ; l++) {
                    if (off_comp(pvect[i+j], pvect[l], offset_thresh)) {
                        last_group = 0 ;
                        break ;
                    }
                }
            }
            if (last_group == 0) {
                for (l=0 ; l<j ; l++) groups[i+l]   = *ngroups + 1 ;
                for (l=0 ; l<k ; l++) groups[i+j+l] = *ngroups + 2 ;
                *ngroups += 2 ;
                i += j+k ;
            } else {
                for (l=0 ; l<j ; l++)               groups[i+l] = *ngroups + 1 ;
                for (l=0 ; l<nraw - (i+j) ; l++)    groups[i+j+l] =*ngroups + 2;
                *ngroups += 2 ;
                i = nraw ;
            }
        }
    }

    /* Groups found ? */
    if (*ngroups == 0) {
        cpl_msg_error(__func__, "Failed to classify in A and B groups") ;
        cpl_free(groups) ;
        return NULL ;
    }

    /* Nb of groups found should be even */
    if (*ngroups % 2) {
        cpl_msg_error(__func__, "Odd number of groups found") ;
        cpl_free(groups) ;
        return NULL ;
    }

    return groups ;
}
    
/*----------------------------------------------------------------------------*/
/**
  @brief    Shift and add groups
  @param    ilist       the input image list
  @param    offsets     the offsets
  @param    groups      the groups ids
  @param    ngroups     the number of groups
  @param    abba_off    the output list offsets
  @return   the new abba image list
 
  The frames in each group are combined together. In principle, there
  should be no offset difference between the frames in the same group,
  so an average should be enough. But to support more general cases
  (generic offset) a shift and add is applied if the offsets differ.

  Example 
  frame1 -  cat A    -> group 1  ---> shift and add 1
  frame2 -  cat A    -> group 1
  frame3 -  cat B    -> group 2  ---> shift and add 2
  frame4 -  cat B    -> group 2
  frame5 -  cat B    -> group 3  ---> shift and add 3
  frame6 -  cat B    -> group 3
  frame7 -  cat A    -> group 4  ---> shift and add 4
  frame8 -  cat A    -> group 4
  frame9 -  cat A    -> group 5  ---> shift and add 5
  frame10 - cat A    -> group 5
  frame11 - cat B    -> group 6  ---> shift and add 6
  frame12 - cat B    -> group 6
  frame13 - cat B    -> group 6
  frame14 - cat B    -> group 6
 */
/*----------------------------------------------------------------------------*/
static cpl_imagelist * crires_util_saa_groups(
        const cpl_imagelist     *   ilist,
        const cpl_vector        *   offsets,
        const int               *   groups,
        int                         ngroups,
        cpl_vector              **  abba_off)
{
    cpl_imagelist       *   abba ; 
    cpl_imagelist       *   group_list ; 
    cpl_image           *   tmp_ima ;
    cpl_image           **  combined ;
    cpl_bivector        *   group_off ;
    double              *   pgroup_off ;
    const double        *   poffsets ;
    double              *   pabba_off ;
    int                     nima ;
    int                     saa ;
    int                     i, j, k ;

    /* Test entries */
    if ((ilist == NULL) || (offsets == NULL) || (groups == NULL)) return NULL ;
    
    /* Initialise */
    nima = cpl_imagelist_get_size(ilist) ;
    poffsets = cpl_vector_get_data_const(offsets) ;

    /* Create the output image list */
    abba = cpl_imagelist_new() ;
    *abba_off = cpl_vector_new(ngroups) ;
    pabba_off = cpl_vector_get_data(*abba_off) ;
    
    /* Loop on the groups */
    for (i=0 ; i<ngroups ; i++) {
        /* Initialise */
        saa = 0 ;
        /* Create the group list of images */
        group_list = cpl_imagelist_new() ;
        k = 0 ;
        for (j=0 ; j<nima ; j++) {
            if (i+1 == groups[j]) {
                /* Get the first offset of the group in abba_off */
                if (k==0) pabba_off[i] = poffsets[j] ;
                /* To know if we need the saa (shift and add) */
                if (fabs(pabba_off[i]-poffsets[j]) > 1e-3) saa = 1 ;
                /* Copy the images of the group in group_list */
                tmp_ima = cpl_image_duplicate(cpl_imagelist_get_const(ilist,j));
                cpl_imagelist_set(group_list, tmp_ima, k) ;
                tmp_ima = NULL ;
                k++ ;
            }
        }

        if (saa) {
            /* Get the offsets of the group in group_off */
            group_off = cpl_bivector_new(k) ;
            cpl_vector_fill(cpl_bivector_get_x(group_off), 0.0) ;
            pgroup_off = cpl_bivector_get_y_data(group_off) ;
            k = 0 ;
            for (j=0 ; j<nima ; j++) {
                if (i+1 == groups[j]) {
                    pgroup_off[k] = poffsets[j] ;
                    k++ ;
                }
            }
            cpl_vector_subtract_scalar(cpl_bivector_get_y(group_off), 
                    pabba_off[i]) ;
            /* Shift and add */
            cpl_msg_debug(__func__, "Apply shift-and-add for group %d", i+1) ;
            if ((combined = cpl_geom_img_offset_saa(group_list,
                           group_off, CPL_KERNEL_DEFAULT, 0, 0,
                           CPL_GEOM_FIRST, NULL, NULL)) == NULL) {
                cpl_msg_error(__func__,"Cannot shift and add group nb %d",i+1) ;
                cpl_imagelist_delete(group_list) ;
                cpl_bivector_delete(group_off) ;
                cpl_imagelist_delete(abba) ;
                cpl_vector_delete(*abba_off) ;
                return NULL ;
            }
            cpl_bivector_delete(group_off) ;
            cpl_image_delete(combined[1]) ;
            cpl_imagelist_set(abba, combined[0], i) ;
            cpl_free(combined) ;
        } else {
            /* Averaging */
            cpl_msg_debug(__func__, "Apply averaging for group %d", i+1) ;
            if ((tmp_ima = cpl_imagelist_collapse_create(group_list)) == NULL) {
                cpl_msg_error(__func__, "Cannot average group nb %d", i+1) ;
                cpl_imagelist_delete(group_list) ;
                cpl_imagelist_delete(abba) ;
                cpl_vector_delete(*abba_off) ;
                return NULL ;
            }
            cpl_imagelist_set(abba, tmp_ima, i) ;
        }
        cpl_imagelist_delete(group_list) ;
    }
    return abba ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Create the nodded images
  @param    abba        the abba images list
  @param    abba_off    the abba offsets
  @param    nodded_off  the offsets of the output image list
  @return   the nodded images

  The frames are reduced by pairs. 
  At this stage, we have an even number of frames.

  Each pair (frame n, frame n+1) will generate two nodded frames 
  (frame n-frame n+1) and (frame n+1-frame n).
  Example 
  abba1 - cat A  ---> abba1 - abba2 = nodded1
  abba2 - cat B  ---> abba2 - abba1 = nodded2
  abba3 - cat B  ---> abba3 - abba4 = nodded3
  abba4 - cat A  ---> abba4 - abba3 = nodded4
  abba5 - cat A  ---> abba5 - abba6 = nodded5
  abba6 - cat B  ---> abba6 - abba5 = nodded6
 */
/*----------------------------------------------------------------------------*/
static cpl_imagelist * crires_util_nodded(
        cpl_imagelist   *   abba,
        cpl_vector      *   abba_off,
        cpl_vector      **  nodded_off)
{
    cpl_imagelist       *   nodded ;
    cpl_image           *   tmp_ima ;
    int                     nima ;
    double                  ref_off ;
    int                     i ;

    /* Test entries */
    if ((abba == NULL) || (abba_off == NULL)) return NULL ;

    /* Initialise */
    nima = cpl_imagelist_get_size(abba) ;
    if (nima % 2) {
        cpl_msg_error(__func__, "Number of images should be even") ;
        return NULL ;
    }
   
    /* Create the offsets between the nodded images */
    *nodded_off = cpl_vector_duplicate(abba_off) ;
    /* The image list to contain the nodded images */
    nodded = cpl_imagelist_new() ;
    for (i=0 ; i<(nima/2) ; i++) {
        /* a-b */
        tmp_ima = cpl_image_duplicate(cpl_imagelist_get(abba, 2*i)) ;
        cpl_image_subtract(tmp_ima, cpl_imagelist_get(abba, 2*i+1)) ;
        cpl_imagelist_set(nodded, tmp_ima, 2*i) ;
        /* b-a */
        tmp_ima = cpl_image_duplicate(cpl_imagelist_get(abba, 2*i+1)) ;
        cpl_image_subtract(tmp_ima, cpl_imagelist_get(abba, 2*i)) ;
        cpl_imagelist_set(nodded, tmp_ima, 2*i+1) ;
    }

    /* Subtract the first offset to the others */
    ref_off = cpl_vector_get(*nodded_off, 0) ;
    cpl_vector_subtract_scalar(*nodded_off, ref_off) ;
    return nodded ;
}


/*----------------------------------------------------------------------------*/
/**
  @brief    Refine the offsets
  @param    ima1    the first image
  @param    ima2    the second image
  @return   the computed offset or 10000.0 in error case
 */
/*----------------------------------------------------------------------------*/
static double crires_util_refine_offset(
        const cpl_image *   ima1,
        const cpl_image *   ima2)
{
    double                  pos1, pos2 ;

    /* Test entries */
    if (ima1 == NULL) return 10000.0 ;
    if (ima2 == NULL) return 10000.0 ;
    
    /* Detect the spectra */
    if (irplib_spectrum_find_brightest(ima1, 0.0, NO_SHADOW, 0.0, 0, 
                &pos1) == -1){
        return 10000.0 ;
    }
    if (irplib_spectrum_find_brightest(ima2, 0.0, NO_SHADOW, 0.0, 0,
                &pos2) == -1){
        return 10000.0 ;
    }
    return pos1-pos2 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    compares two doubles in regard to a given threshold
  @param    off1 first double
  @param    off2 second double
  @param    thresh  threshold
  @return   1 if threshold is between offsets, 0 otherwise
 */
/*----------------------------------------------------------------------------*/
static int off_comp(double off1, double off2, double thresh)
{
    if (((off1>thresh) && (off2<thresh)) || ((off1<thresh) && (off2>thresh)))
        return 1 ;
    else return 0 ;
}

