/* $Id: crires_wlestimate.c,v 1.9 2011-11-22 08:50:12 yjung Exp $
 *
 * This file is part of the CRIRES Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: yjung $
 * $Date: 2011-11-22 08:50:12 $
 * $Revision: 1.9 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include <math.h>
#include <cpl.h>

#include "crires_wlestimate.h"
#include "crires_pfits.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup crires_wlestimate  Wavelength calibration estimation
 *
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Getthe Wavelength Calibration estimate
  @param    fname       the file name for the configuration
  @param    chip        the chip to calibrate (1 to 4)
  @return   the polynomial or NULL in error case

  The returned polynomial must be deallocated with cpl_polynomial_delete()
 */
/*----------------------------------------------------------------------------*/
cpl_polynomial * crires_wlestimate_get(
        const char  *   fname,
        int             chip)
{
    cpl_propertylist    *   plist ;
    double                  wmin, wmax ;

    /* Test entries */
    if (chip < 1 || chip >4) return NULL ;
    if (fname == NULL) return NULL ;
    
    /* Initialise */
    wmin = wmax = -1.0 ;

    /* Get the wavelengths */
    if (cpl_error_get_code() != CPL_ERROR_NONE) return NULL ;
    if ((plist = cpl_propertylist_load(fname, 0)) == NULL) {
        cpl_msg_error(__func__, "getting header");
        return NULL ;
    }
    
    /* Check the chip */
    if (chip == 1)      { 
        wmin = crires_pfits_get_wlens1(plist) ;
        wmax = crires_pfits_get_wlene1(plist) ;
    } else if (chip == 2) { 
        wmin = crires_pfits_get_wlens2(plist) ;
        wmax = crires_pfits_get_wlene2(plist) ;
    } else if (chip == 3) { 
        wmin = crires_pfits_get_wlens3(plist) ;
        wmax = crires_pfits_get_wlene3(plist) ;
    } else if (chip == 4) { 
        wmin = crires_pfits_get_wlens4(plist) ;
        wmax = crires_pfits_get_wlene4(plist) ;
    }
    cpl_propertylist_delete(plist) ;
    
    /* Try in extensions headers */
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_info(__func__, "Try to find the information in the extension") ;
        cpl_error_reset() ;
        if ((plist = cpl_propertylist_load(fname, chip)) == NULL) {
            cpl_msg_error(__func__, "getting extension header");
            return NULL ;
        }
        wmin = crires_pfits_get_wlens(plist) ;
        wmax = crires_pfits_get_wlene(plist) ;
        cpl_propertylist_delete(plist) ;
        if (cpl_error_get_code() != CPL_ERROR_NONE) return NULL ;
    }

    if (wmin <= 0.0 || wmax <= 0.0) return NULL ;
    
    cpl_msg_info(__func__, "Wavelength range: %g - %g", wmin, wmax) ;
   
    return crires_wlestimate_compute(wmin, wmax) ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the polynomial from boundaries
  @param    wmin    First pixel wavelength
  @param    wmax    Last pixel wavelength
  @return   the polynomial or NULL in error case

  The returned polynomial must be deallocated with cpl_polynomial_delete()
 */
/*----------------------------------------------------------------------------*/
cpl_polynomial * crires_wlestimate_compute(
        double          wmin,
        double          wmax)
{
    cpl_polynomial      *   poly ;
    cpl_size                power ;
    double                  a, b ;
    int                     nbpix = 1024 ;

    /* Test entries */
    if (wmin <= 0.0) return NULL ;
    if (wmax <= 0.0) return NULL ;
    if (wmax <= wmin) return NULL ;
    
    /* Compute polynomial coeffs */
    a = (wmax - wmin) / (nbpix-1) ;
    b = wmin - a ;

    /* Create polynomial */
    poly = cpl_polynomial_new(1) ;
    power = 0 ;
    cpl_polynomial_set_coeff(poly, &power, b) ;
    power = 1 ;
    cpl_polynomial_set_coeff(poly, &power, a) ;
    
    return poly ;
}

/**@}*/
