/* $Id: crires_util_genconfig.c,v 1.19 2011-03-22 09:17:12 yjung Exp $
 *
 * This file is part of the CRIRES Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: yjung $
 * $Date: 2011-03-22 09:17:12 $
 * $Revision: 1.19 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <locale.h>
#include "crires_recipe.h"

/*-----------------------------------------------------------------------------
                                Define
 -----------------------------------------------------------------------------*/

#define RECIPE_STRING "crires_util_genconfig"

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int crires_util_genconfig_save(cpl_table *, const cpl_parameterlist *, 
        cpl_frameset *);

static char crires_util_genconfig_description[] =
"This recipe is used to generate the model configuration file.\n"
"The sof file contains the names of the input ASCII file\n"
"tagged with "CRIRES_UTIL_GENCONFIG_RAW".\n"
"The ASCII file must contain five columns:\n"
"1st: The best guess value\n"
"2nd: The low limit\n"
"3th: The high limit\n"
"4th: The Flag to recompute or not\n"
"5th: Name of the Parameter\n"
"The ASCII files are in the catalogs/ directory of the CRIRES distribution.\n"
"This recipe produces 1 file:\n"
"First product:     the table with the configuration for the model.\n"
"                   (PRO TYPE = "CRIRES_PROTYPE_MOD_CONF")\n" ;

CRIRES_RECIPE_DEFINE(crires_util_genconfig,
        CRIRES_PARAM_CONFIG_MODE,
        "Generate spectrum calibration FITS tables",
        crires_util_genconfig_description) ;

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Input */
    int         mode ; 
} crires_util_genconfig_config ;

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    The FITS file creation occurs here 
  @param    framelist   the frames list
  @return   0 iff everything is ok
  The recipe expects a text file and will create a FITS file out of it. 
 */
/*----------------------------------------------------------------------------*/
static int crires_util_genconfig(
        cpl_frameset            *   framelist,
        const cpl_parameterlist *   parlist)
{
    FILE            *   in ;
    char                line[1024];
    cpl_frame       *   cur_frame ;
    const char      *   cur_fname ;
    int                 nentries ;
    char                name[1024] ;
    double              best, low, high ;
    int                 flag ;
    cpl_table       *   tab ;
    int                 i ;

    /* Needed for sscanf() */
    setlocale(LC_NUMERIC, "C");

    /* Retrieve input parameters */
    crires_util_genconfig_config.mode = crires_parameterlist_get_int(parlist,
            RECIPE_STRING, CRIRES_PARAM_CONFIG_MODE) ;

    /* Identify the RAW and CALIB frames in the input frameset */
    if (crires_dfs_set_groups(framelist, NULL)) {
        cpl_msg_error(__func__, "Cannot identify RAW and CALIB frames") ;
        return -1 ;
    }
  
    /* Get the config file name */
    cur_frame = cpl_frameset_get_position(framelist, 0) ;
    cur_fname = cpl_frame_get_filename(cur_frame) ;

    /* Open the file */
    if ((in = fopen(cur_fname, "r")) == NULL) {
        cpl_msg_error(__func__, "Could not open %s", cur_fname) ;
        return -1 ;
    }

    /* Count number of entries */
    nentries = 0 ;
    while (fgets(line, 1024, in) != NULL) {
        if (line[0] != '#' && sscanf(line, "%lg %lg %lg %d %s", 
                    &best, &low, &high, &flag, name) == 5)  nentries++ ;
    }
    if (nentries == 0) {
        cpl_msg_error(__func__, "No valid entryin the file") ;
        fclose(in) ;
        return -1 ;
    }
        
    /* Create the output table */
    tab = cpl_table_new(nentries) ;
    cpl_table_new_column(tab, CRIRES_COL_MODEL_CONF_BEST, CPL_TYPE_DOUBLE) ;
    cpl_table_new_column(tab, CRIRES_COL_MODEL_CONF_LOW, CPL_TYPE_DOUBLE) ;
    cpl_table_new_column(tab, CRIRES_COL_MODEL_CONF_HIGH, CPL_TYPE_DOUBLE) ;
    cpl_table_new_column(tab, CRIRES_COL_MODEL_CONF_FLAG, CPL_TYPE_INT) ;
    cpl_table_new_column(tab, CRIRES_COL_MODEL_CONF_NAME, CPL_TYPE_STRING) ;
    
    /* Fill the table */
    i = 0 ;
    rewind(in) ;
    while (fgets(line, 1024, in) != NULL) {
        if (line[0] != '#' && sscanf(line, "%lg %lg %lg %d %s", 
                    &best, &low, &high, &flag, name) == 5) {
            cpl_table_set_string(tab, CRIRES_COL_MODEL_CONF_NAME, i, name) ;
            cpl_table_set_double(tab, CRIRES_COL_MODEL_CONF_BEST, i, best) ;
            cpl_table_set_double(tab, CRIRES_COL_MODEL_CONF_LOW, i, low) ;
            cpl_table_set_double(tab, CRIRES_COL_MODEL_CONF_HIGH, i, high) ;
            cpl_table_set_int(tab, CRIRES_COL_MODEL_CONF_FLAG, i, flag) ;
            i++ ;
        }
    }
    fclose(in) ;
   
    /* Save the table */
    cpl_msg_info(__func__, "Saving the table with %d rows", nentries) ;
    if (crires_util_genconfig_save(tab, parlist, framelist) == -1) {
        cpl_msg_error(__func__, "Cannot write the table") ;
        cpl_table_delete(tab) ;
        return -1 ;
    }
    cpl_table_delete(tab) ;
    return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save the product of the recipe
  @param    out_table   the table 
  @param    parlist     the input list of parameters
  @param    set         the input frame set
  @return   0 if everything is ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
static int crires_util_genconfig_save(
        cpl_table               *   out_table,
        const cpl_parameterlist *   parlist,
        cpl_frameset            *   set)
{
    cpl_propertylist    *   plist ;
    const char          *   procat ;

    if (crires_util_genconfig_config.mode == 1) 
        procat = CRIRES_CALPRO_MODEL_REFINE_CONF ;
    else if (crires_util_genconfig_config.mode == 2)
        procat = CRIRES_CALPRO_MODEL_CONFIG ;
    else {
        cpl_msg_error(__func__, "Unsupported mode") ;
        return -1 ;
    }

    plist = cpl_propertylist_new();
    cpl_propertylist_append_string(plist, "INSTRUME", "CRIRES") ;
    cpl_propertylist_append_string(plist, CPL_DFS_PRO_CATG, procat) ;
    cpl_propertylist_append_string(plist, CPL_DFS_PRO_TYPE,
            CRIRES_PROTYPE_MOD_CONF) ;
    cpl_propertylist_append_string(plist, "ESO PRO CFG-IDENT",
            "AlwaysOK") ;

    if (cpl_dfs_save_table(set, NULL, parlist, set, NULL, out_table,
                NULL, "crires_util_genconfig", plist, NULL, 
                PACKAGE "/" PACKAGE_VERSION,
                "crires_util_genconfig.fits") != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "Cannot save the table") ;
        return -1 ;
    }
    cpl_propertylist_delete(plist) ;
    
    /* Return */
    return 0 ;
}
