/* $Id: crires_util_genlines.c,v 1.16 2011-11-24 08:27:46 yjung Exp $
 *
 * This file is part of the CRIRES Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: yjung $
 * $Date: 2011-11-24 08:27:46 $
 * $Revision: 1.16 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include "crires_recipe.h"

/*-----------------------------------------------------------------------------
                                Define
 -----------------------------------------------------------------------------*/

#define RECIPE_STRING "crires_util_genlines"

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int crires_util_genlines_save(cpl_table *, const cpl_parameterlist *, 
        cpl_frameset *);

static char crires_util_genlines_description[] =
"This recipe is used to generate spectrum calibration tables.\n"
"The sof file contains the names of the input ASCII file\n"
"tagged with "CRIRES_UTIL_GENLINES_RAW".\n"
"The ASCII file must contain two columns:\n"
"1st: Wavelengths in increasing order (the unit is corrected by\n"
"     the factor option to obtain nanometers).\n"
"2nd: The atmospheric emission.\n"
"The ASCII files are in the catalogs/ directory of the CRIRES distribution.\n"
"This recipe produces 1 file:\n"
"First product:     the table with the lines.\n"
"                   (PRO TYPE = "CRIRES_PROTYPE_CATALOG")\n" ;

CRIRES_RECIPE_DEFINE(crires_util_genlines,
        CRIRES_PARAM_FILL_BLANKS    |
        CRIRES_PARAM_PLOT           |
        CRIRES_PARAM_LINES_MODE     |
        CRIRES_PARAM_WL_FACTOR,
        "Generate spectrum calibration FITS tables",
        crires_util_genlines_description) ;

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    int             fill_blanks ; 
    int             display ;
    int             mode ;
    double          wl_factor ;
    /* Outputs */
} crires_util_genlines_config ;

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    The FITS file creation occurs here 
  @param    framelist   the frames list
  @return   0 iff everything is ok

  The recipe expects a text file with two columns, and will create a FITS file
  out of it. 
 */
/*----------------------------------------------------------------------------*/
static int crires_util_genlines(
        cpl_frameset            *   framelist,
        const cpl_parameterlist *   parlist)
{
    cpl_bivector    *   bivec ;
    double          *   pbivec_x ;
    double          *   pbivec_y ;
    cpl_bivector    *   bivec_fill ;
    double          *   pbivec_fill_x ;
    double          *   pbivec_fill_y ;
    cpl_frame       *   cur_frame ;
    int                 nvals, nb_new_vals ;
    double              wavel ;
    cpl_table       *   tab ;
    int                 i ;

    /* Retrieve input parameters */
    crires_util_genlines_config.fill_blanks = crires_parameterlist_get_bool(
            parlist, RECIPE_STRING, CRIRES_PARAM_FILL_BLANKS) ;
    crires_util_genlines_config.display = crires_parameterlist_get_bool(
                        parlist, RECIPE_STRING, CRIRES_PARAM_PLOT) ;
    crires_util_genlines_config.mode = crires_parameterlist_get_int(
            parlist, RECIPE_STRING, CRIRES_PARAM_LINES_MODE) ;
    crires_util_genlines_config.wl_factor = crires_parameterlist_get_double( 
            parlist, RECIPE_STRING, CRIRES_PARAM_WL_FACTOR) ;
   
    /* Identify the RAW and CALIB frames in the input frameset */
    if (crires_dfs_set_groups(framelist, NULL)) {
        cpl_msg_error(__func__, "Cannot identify RAW and CALIB frames") ;
        return -1 ;
    }
   
    /* Load the file */
    cur_frame = cpl_frameset_get_position(framelist, 0) ;
    if ((bivec=cpl_bivector_read(cpl_frame_get_filename(cur_frame)))==NULL) {
        cpl_msg_error(__func__, "Cannot load the file in the bivector") ;
        return -1 ;
    }
    nvals = cpl_bivector_get_size(bivec) ;
    
     /* If fill_blanks is requested */
    if (crires_util_genlines_config.fill_blanks) {
        nb_new_vals = 3 * nvals ;
        bivec_fill = cpl_bivector_new(nb_new_vals) ;
        pbivec_fill_x = cpl_bivector_get_x_data(bivec_fill) ;
        pbivec_fill_y = cpl_bivector_get_y_data(bivec_fill) ;
        pbivec_x = cpl_bivector_get_x_data(bivec) ;
        pbivec_y = cpl_bivector_get_y_data(bivec) ;
        for (i=0 ; i<nvals ; i++) {
            wavel = pbivec_x[i] * crires_util_genlines_config.wl_factor ;
            pbivec_fill_x[3*i] = wavel - 0.01 ;
            pbivec_fill_y[3*i] = 0.0 ;
            pbivec_fill_x[3*i+1] = wavel ;
            pbivec_fill_y[3*i+1] = pbivec_y[i] ;
            pbivec_fill_x[3*i+2] = wavel + 0.01 ;
            pbivec_fill_y[3*i+2] = 0.0 ;
        }
        cpl_bivector_delete(bivec);
        bivec = bivec_fill ;
        bivec_fill = NULL ;
        nvals = cpl_bivector_get_size(bivec) ;
    } else {
        cpl_vector_multiply_scalar(cpl_bivector_get_x(bivec),
                crires_util_genlines_config.wl_factor) ;
    }

    /* Display if requested */
    if (crires_util_genlines_config.display) {
        cpl_plot_bivector(
                "set grid;set xlabel 'Wavelength (nm)';set ylabel 'Emission';",
                "t 'Catalog lines' w lines", "", bivec);
    }
    
    /* Allocate the data container */
    tab = cpl_table_new(nvals) ;
    cpl_table_wrap_double(tab, cpl_bivector_get_x_data(bivec),
            CRIRES_COL_WAVELENGTH) ;
    cpl_table_wrap_double(tab, cpl_bivector_get_y_data(bivec), 
            CRIRES_COL_EMISSION) ;
   
    /* Save the table */
    cpl_msg_info(__func__, "Saving the table with %d rows", nvals) ;
    if (crires_util_genlines_save(tab, parlist, framelist) == -1) {
        cpl_msg_error(__func__, "Cannot write the table") ;
        cpl_bivector_delete(bivec) ;
        cpl_table_unwrap(tab, CRIRES_COL_WAVELENGTH) ;
        cpl_table_unwrap(tab, CRIRES_COL_EMISSION) ;
        cpl_table_delete(tab) ;
        return -1 ;
    }
    cpl_bivector_delete(bivec) ;
    cpl_table_unwrap(tab, CRIRES_COL_WAVELENGTH) ;
    cpl_table_unwrap(tab, CRIRES_COL_EMISSION) ;
    cpl_table_delete(tab) ;
    return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save the product of the recipe
  @param    out_table   the table 
  @param    parlist     the input list of parameters
  @param    set         the input frame set
  @return   0 if everything is ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
static int crires_util_genlines_save(
        cpl_table               *   out_table,
        const cpl_parameterlist *   parlist,
        cpl_frameset            *   set)
{
    cpl_propertylist    *   plist ;
    const char          *   procat ;

    if (crires_util_genlines_config.mode == 1) 
        procat = CRIRES_CALPRO_HITRAN_CAT ;
    else if (crires_util_genlines_config.mode == 2)
        procat = CRIRES_CALPRO_OH_CAT ;
    else if (crires_util_genlines_config.mode == 3)
        procat = CRIRES_CALPRO_THAR_CAT ;
    else if (crires_util_genlines_config.mode == 4)
        procat = CRIRES_CALPRO_N2O_CAT ;
    else if (crires_util_genlines_config.mode == 5)
        procat = CRIRES_CALPRO_MODEL_WAVEEG ;
    else 
        procat = "UNKNOWN" ;

    plist = cpl_propertylist_new();
    cpl_propertylist_append_string(plist, "INSTRUME", "CRIRES") ;
    cpl_propertylist_append_string(plist, CPL_DFS_PRO_CATG, procat) ;
    cpl_propertylist_append_string(plist, CPL_DFS_PRO_TYPE, 
            CRIRES_PROTYPE_CATALOG) ;

    if (cpl_dfs_save_table(set, NULL, parlist, set, NULL, out_table,
                NULL, "crires_util_genlines", plist, NULL,
                PACKAGE "/" PACKAGE_VERSION,
                "crires_util_genlines.fits") != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "Cannot save the table") ;
        return -1 ;
    }
    cpl_propertylist_delete(plist) ;

    /* Return */
    return 0 ;
}
